/*
 * $Id: keyex.c,v 1.1 2005/01/05 14:06:02 pickett Exp $
 *
 * AOChat -- library for talking with the Anarchy Online chat servers
 * Copyright (C) 2002, 2003  Oskari Saarenmaa <auno@auno.org>.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *
 * ***
 *
 * The license does not actually allow this program to be linked against
 * PHP, which uses a different license, but since I own the copyrights
 * to this code, I hereby explicitly allow linking this object to a PHP
 * module called 'aokex' which is also authored by me.
 *
 * Oskari Saarenmaa / auno@auno.org / 2005-Jan-05.
 *
 *
 */

/*
 * based on the PHP class I wrote some time ago.. this C-port is
 * quite unfinished, the only thing it can do at the moment is generate
 * the required login keys (onesided Diffie-Hellman key exchange with
 * TEA encrypted message body)
 *
 * DH:  http://www.rsasecurity.com/rsalabs/faq/3-6-1.html
 * TEA: http://www.ftp.cl.cam.ac.uk/ftp/papers/djw-rmn/djw-rmn-tea.html
 *
 */


#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>

#include "aokex.h"

static char * aochat_crypt(char *, char *, int);
static void   aochat_tea_encrypt(AoUInt32[4], AoUInt32[4]);


char *
generate_login_key(char *servkey, char *username, char *password)
{
  gmp_randstate_t rstate;
  mpz_t dh_p, dh_y, dh_g, dh_k, dh_x, dh_secret, tmp;
  char *dh_str_x, *dh_str_k;
  char *challenge, *plaintext, *ciphertext;
  size_t clen, plen;
  AoUInt32 rnum;
  
  const char
    dh_str_y[] =
    "9c32cc23d559ca90fc31be72df817d0e124769e809f936bc"
    "14360ff4bed758f260a0d596584eacbbc2b88bdd41041616"
    "3e11dbf62173393fbc0c6fefb2d855f1a03dec8e9f105bba"
    "d91b3437d8eb73fe2f44159597aa4053cf788d2f9d7012fb"
    "8d7c4ce3876f7d6cd5d0c31754f4cd96166708641958de54"
    "a6def5657b9f2e92",
    dh_str_p[] =
    "eca2e8c85d863dcdc26a429a71a9815ad052f6139669dd65"
    "9f98ae159d313d13c6bf2838e10a69b6478b64a24bd054ba"
    "8248e8fa778703b418408249440b2c1edd28853e240d8a7e"
    "49540b76d120d3b1ad2878b1b99490eb4a2a5e84caa8a91c"
    "ecbdb1aa7c816e8be343246f80c637abc653b893fd91686c"
    "f8d32d6cfe5f2a6f",
    dh_str_g[] = "5";
  const int
    dh_exponent = 192;
  
  gmp_randinit_default(rstate);
  gmp_randseed_ui(

  /* initialize constants, prime, generator and funcom's Y */
  mpz_init_set_str(dh_y, dh_str_y, 16);
  mpz_init_set_str(dh_p, dh_str_p, 16);
  mpz_init_set_str(dh_g, dh_str_g, 16);

  /* generate our secret key, we use 192 bits of entropy */
  mpz_init(dh_secret);
  mpz_urandomb(dh_secret, rstate, dh_exponent);
  /* gen^secret mod prime, our 'X' */
  mpz_init(dh_x);
  mpz_powm(dh_x, dh_g, dh_secret, dh_p);
  gmp_asprintf(&dh_str_x, "%Zx", dh_x);

  /* calculate the session key */
  mpz_init(dh_k);
  mpz_powm(dh_k, dh_y, dh_secret, dh_p);
  gmp_asprintf(&dh_str_k, "%Zx", dh_k);
  
  mpz_clear(dh_k);
  mpz_clear(dh_x);
  mpz_clear(dh_secret);
  mpz_clear(dh_p);
  mpz_clear(dh_y);
  mpz_clear(dh_g);
  
  /* challenge = username '|' servkey '|' password */
  /* payload = cookie[8] challenge-length[4] challenge */
  clen = strlen(username) + 1 + strlen(servkey) + 1 + strlen(password);
  plen = 8+4+clen;
  plen = ((plen+7)/8)*8; /* round up to 8 */
  
  plaintext = (char *)malloc(plen+1); /* make room for a null */
  
  printf("clen: %d, plen: %d, plaintext: %08x\n", clen, plen, plaintext);
  
  /* generate the cookie */
  mpz_init(tmp);
  mpz_urandomb(tmp, rstate, 32);
  rnum = (AoUInt32)mpz_get_ui(tmp);
  memcpy(plaintext+0, &rnum, sizeof(AoUInt32));
  mpz_urandomb(tmp, rstate, 32);
  rnum = (AoUInt32)mpz_get_ui(tmp);
  memcpy(plaintext+4, &rnum, sizeof(AoUInt32));
  mpz_clear(tmp);
  gmp_randclear(rstate);

  /* pack the content-length in network byte order */
  plaintext[ 8] = (clen>>24) & 0xff;
  plaintext[ 9] = (clen>>16) & 0xff;
  plaintext[10] = (clen>> 8) & 0xff;
  plaintext[11] = (clen>> 0) & 0xff;

  /* insert the challenge in our plaintext string, plus padding */
  snprintf(plaintext+12, plen-12, "%s|%s|%s                ",
    username, servkey, password);

#if 1
  plaintext[plen] = 0;
  return plaintext;
}
int main()
{
  int i;
  char *text = generate_login_key("asdf", "pickett", "kala");
  for(i=0; i<32;i++)
    putc(text[i], stdout);
  putc('\n', stdout);
  return 0;
}
#else

  /* 'crypted' is a malloc()ed string from aochat_crypt() */
  crypted = aochat_crypt(dh_str_secret, plain, tlen);
  
  /* store the data in the pointer provided to us */
  *retkey = (char*)malloc(strlen(dh_str_k)+tlen*2+2);
  sprintf(*retkey, "%s-%s", dh_str_k, crypted);

  /* free any memory we allocated */
  free(crypted);
  free(plain);
  free(dh_str_k);
}

static char *
aochat_crypt(char *key, char *str, int len)
{ 
  AoUInt32 i, cycle[4], akey[4];
  char *crypted, *p;
  
  /* Our return buffer, and pointer to it */
  p = crypted = (char*)malloc(len*2+1);

  /* Zero the cycle */
  memset(cycle, 0, sizeof cycle);

  /* Convert the hexadecimal key to binary */
  for(i=0; i<16; i++)
    *((char*)&akey+i) = htoi(key[2*i])*16 + htoi(key[2*i+1]);

  for(i=0; i<len/4;)
  { 
    cycle[0] = (*(AoUInt32*)(str+4*i++)) ^ (cycle[2]);
    cycle[1] = (*(AoUInt32*)(str+4*i++)) ^ (cycle[3]);
    aochat_tea_encrypt(cycle, akey);
    sprintf(p, "%08x%08x", htonl(cycle[2]), htonl(cycle[3]));
    p += 16;
  }
  
  return crypted;
}

static void
aochat_tea_encrypt(AoUInt32 cycle[4], AoUInt32 key[4])
{ 
  AoUInt32 a = cycle[0],
           b = cycle[1],
           sum = 0,
           delta = 0x9e3779b9,
           i = 32;

  while(i--)
  { 
    sum += delta;
    a   += ((b << 4 & 0xfffffff0) + key[0]) ^ (b + sum) ^ ((b >> 5 & 0x7ffffff) + key[1]);
    b   += ((a << 4 & 0xfffffff0) + key[2]) ^ (a + sum) ^ ((a >> 5 & 0x7ffffff) + key[3]);
  }

  cycle[2] = a;
  cycle[3] = b;
}
#endif
